@@ -26,7 +26,7 @@
  * V1_MAX_PAGES fixes, jj, 990325.
  * sparc64 fixes, jj, 000219.
  *
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+ * 1999-02-22 Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  */
@@ -54,9 +54,11 @@
 #include "blkdev.h"
 #include "pathnames.h"
 #include "wholedisk.h"
-#include "writeall.h"
+#include "all-io.h"
 #include "xalloc.h"
 #include "c.h"
+#include "closestream.h"
+#include "ismounted.h"
 
 #ifdef HAVE_LIBUUID
 # include <uuid.h>
@@ -143,21 +145,21 @@ is_sparc64(void)
  * What to do? Let us allow the user to specify the pagesize explicitly.
  *
  */
-static long user_pagesize;
-static int pagesize;
+static unsigned int user_pagesize;
+static unsigned int pagesize;
 static unsigned long *signature_page = NULL;
 
 static void
 init_signature_page(void)
 {
 
-	int kernel_pagesize = pagesize = getpagesize();
+	unsigned int kernel_pagesize = pagesize = getpagesize();
 
 	if (user_pagesize) {
-		if ((user_pagesize & (user_pagesize - 1)) ||
-		    user_pagesize < (long) sizeof(struct swap_header_v1_2) + 10)
+		if (!is_power_of_2(user_pagesize) ||
+		    user_pagesize < sizeof(struct swap_header_v1_2) + 10)
 			errx(EXIT_FAILURE,
-				_("Bad user-specified page size %lu"),
+				_("Bad user-specified page size %u"),
 				user_pagesize);
 		pagesize = user_pagesize;
 	}
@@ -314,17 +316,20 @@ check_blocks(void)
 	buffer = xmalloc(pagesize);
 	current_page = 0;
 	while (current_page < PAGES) {
+
+		ssize_t rc;
+
 		if (do_seek && lseek(DEV,current_page*pagesize,SEEK_SET) !=
 		    current_page*pagesize)
 			errx(EXIT_FAILURE, _("seek failed in check_blocks"));
-		if ((do_seek = (pagesize != read(DEV, buffer, pagesize))))
+
+		rc = read(DEV, buffer, pagesize);
+		do_seek = (rc < 0 || (size_t) rc != pagesize);
+		if (do_seek)
 			page_bad(current_page);
 		current_page++;
 	}
-	if (badpages == 1)
-		printf(_("one bad page\n"));
-	else if (badpages > 1)
-		printf(_("%lu bad pages\n"), badpages);
+	printf(P_("%lu bad page\n", "%lu bad pages\n", badpages), badpages);
 	free(buffer);
 }
 
@@ -347,29 +352,6 @@ get_size(const char *file)
 	return size;
 }
 
-/*
- * Check to make certain that our new filesystem won't be created on
- * an already mounted partition.  Code adapted from mke2fs, Copyright
- * (C) 1994 Theodore Ts'o.  Also licensed under GPL.
- * (C) 2006 Karel Zak -- port to mkswap
- */
-static int
-check_mount(void)
-{
-	FILE *f;
-	struct mntent *mnt;
-
-	if ((f = setmntent (_PATH_MOUNTED, "r")) == NULL)
-		return 0;
-	while ((mnt = getmntent (f)) != NULL)
-		if (strcmp (device_name, mnt->mnt_fsname) == 0)
-			break;
-	endmntent (f);
-	if (!mnt)
-		return 0;
-	return 1;
-}
-
 #ifdef HAVE_LIBBLKID
 static blkid_probe
 new_prober(int fd)
@@ -384,10 +366,9 @@ new_prober(int fd)
 #endif
 
 static void
-wipe_device(int fd, const char *devname, int force, int is_blkdev)
+wipe_device(int fd, const char *devname, int force)
 {
 	char *type = NULL;
-	int whole = 0;
 	int zap = 1;
 #ifdef HAVE_LIBBLKID
 	blkid_probe pr = NULL;
@@ -396,29 +377,21 @@ wipe_device(int fd, const char *devname, int force, int is_blkdev)
 		if (lseek(fd, 0, SEEK_SET) != 0)
 			errx(EXIT_FAILURE, _("unable to rewind swap-device"));
 
-		if (is_blkdev && is_whole_disk_fd(fd, devname)) {
-			/* don't zap bootbits on whole disk -- we know nothing
-			 * about bootloaders on the device */
-			whole = 1;
-			zap = 0;
-		} else {
 #ifdef HAVE_LIBBLKID
-			pr = new_prober(fd);
-			blkid_probe_enable_partitions(pr, 1);
-			blkid_probe_enable_superblocks(pr, 0);
+		pr = new_prober(fd);
+		blkid_probe_enable_partitions(pr, 1);
+		blkid_probe_enable_superblocks(pr, 0);
 
-			if (blkid_do_fullprobe(pr) == 0 &&
-			    blkid_probe_lookup_value(pr, "PTTYPE",
-					(const char **) &type, NULL) == 0 &&
-			    type) {
-				type = xstrdup(type);
-				zap = 0;
-			}
-#else
-			/* don't zap if compiled without libblkid */
+		if (blkid_do_fullprobe(pr) == 0 &&
+		    blkid_probe_lookup_value(pr, "PTTYPE",
+				(const char **) &type, NULL) == 0 && type) {
+			type = xstrdup(type);
 			zap = 0;
-#endif
 		}
+#else
+		/* don't zap if compiled without libblkid */
+		zap = 0;
+#endif
 	}
 
 	if (zap) {
@@ -442,21 +415,24 @@ wipe_device(int fd, const char *devname, int force, int is_blkdev)
 
 		blkid_probe_enable_superblocks(pr, 1);
 		blkid_probe_enable_partitions(pr, 0);
-		blkid_probe_set_superblocks_flags(pr, BLKID_SUBLKS_MAGIC);
+		blkid_probe_set_superblocks_flags(pr, BLKID_SUBLKS_MAGIC|BLKID_SUBLKS_TYPE);
 
-		while (blkid_do_probe(pr) == 0)
+		while (blkid_do_probe(pr) == 0) {
+			const char *data = NULL;
+
+			if (blkid_probe_lookup_value(pr, "TYPE", &data, NULL) == 0 && data)
+				warnx(_("%s: warning: wiping old %s signature."), devname, data);
 			blkid_do_wipe(pr, 0);
+		}
 #endif
 	} else {
 		warnx(_("%s: warning: don't erase bootbits sectors"),
 			devname);
 		if (type)
 			fprintf(stderr, _("        (%s partition table detected). "), type);
-		else if (whole)
-			fprintf(stderr, _("        on whole disk. "));
 		else
 			fprintf(stderr, _("        (compiled without libblkid). "));
-		fprintf(stderr, "Use -f to force.\n");
+		fprintf(stderr, _("Use -f to force.\n"));
 	}
 #ifdef HAVE_LIBBLKID
 	blkid_free_probe(pr);
@@ -473,7 +449,7 @@ main(int argc, char **argv) {
 	unsigned long long sz;
 	off_t offset;
 	int force = 0;
-	long version = 1;
+	int version = 1;
 	char *block_count = 0;
 	char *opt_label = NULL;
 	unsigned char *uuid = NULL;
@@ -496,6 +472,7 @@ main(int argc, char **argv) {
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	while((c = getopt_long(argc, argv, "cfp:L:v:U:Vh", longopts, NULL)) != -1) {
 		switch (c) {
@@ -506,25 +483,24 @@ main(int argc, char **argv) {
 			force=1;
 			break;
 		case 'p':
-			user_pagesize = strtol_or_err(optarg, _("parse page size failed"));
+			user_pagesize = strtou32_or_err(optarg, _("parsing page size failed"));
 			break;
 		case 'L':
 			opt_label = optarg;
 			break;
 		case 'v':
-			version = strtol_or_err(optarg, _("parse version number failed"));
+			version = strtos32_or_err(optarg, _("parsing version number failed"));
 			break;
 		case 'U':
 #ifdef HAVE_LIBUUID
 			opt_uuid = optarg;
 #else
-			warnx(_("warning: ignore -U (UUIDs are unsupported by %s)"),
+			warnx(_("warning: ignoring -U (UUIDs are unsupported by %s)"),
 				program_invocation_short_name);
 #endif
 			break;
 		case 'V':
-			printf(_("%s from %s\n"), program_invocation_short_name,
-						  PACKAGE_STRING);
+			printf(UTIL_LINUX_VERSION);
 			exit(EXIT_SUCCESS);
 		case 'h':
 			usage(stdout);
@@ -537,19 +513,18 @@ main(int argc, char **argv) {
 	if (optind < argc)
 		block_count = argv[optind++];
 	if (optind != argc) {
-		warnx(("only one device as argument is currently supported."));
+		warnx(_("only one device argument is currently supported"));
 		usage(stderr);
 	}
 
 	if (version != 1)
 		errx(EXIT_FAILURE,
-			_("does not support swapspace version %lu."),
-			version);
+			_("swapspace version %d is not supported"), version);
 
 #ifdef HAVE_LIBUUID
 	if(opt_uuid) {
 		if (uuid_parse(opt_uuid, uuid_dat) != 0)
-			errx(EXIT_FAILURE, _("error: UUID parsing failed"));
+			errx(EXIT_FAILURE, _("error: parsing UUID failed"));
 	} else
 		uuid_generate(uuid_dat);
 	uuid = uuid_dat;
@@ -563,12 +538,8 @@ main(int argc, char **argv) {
 	}
 	if (block_count) {
 		/* this silly user specified the number of blocks explicitly */
-		long long blks;
-
-		blks = strtoll_or_err(block_count, "parse block count failed");
-		if (blks < 0)
-			usage(stderr);
-
+		uint64_t blks = strtou64_or_err(block_count,
+					_("invalid block count argument"));
 		PAGES = blks / (pagesize / 1024);
 	}
 	sz = get_size(device_name);
@@ -602,6 +573,11 @@ main(int argc, char **argv) {
 			PAGES * pagesize / 1024);
 	}
 
+	if (is_mounted(device_name))
+		errx(EXIT_FAILURE, _("error: "
+			"%s is mounted; will not make swapspace"),
+			device_name);
+
 	if (stat(device_name, &statbuf) < 0) {
 		perror(device_name);
 		exit(EXIT_FAILURE);
@@ -616,24 +592,15 @@ main(int argc, char **argv) {
 		exit(EXIT_FAILURE);
 	}
 
-	/* Want a block device. Probably not /dev/hda or /dev/hdb. */
 	if (!S_ISBLK(statbuf.st_mode))
 		check=0;
-	else if (statbuf.st_rdev == 0x0300 || statbuf.st_rdev == 0x0340)
-		errx(EXIT_FAILURE, _("error: "
-			"will not try to make swapdevice on '%s'"),
-			device_name);
-	else if (check_mount())
-		errx(EXIT_FAILURE, _("error: "
-			"%s is mounted; will not make swapspace."),
-			device_name);
 	else if (blkdev_is_misaligned(DEV))
 		warnx(_("warning: %s is misaligned"), device_name);
 
 	if (check)
 		check_blocks();
 
-	wipe_device(DEV, device_name, force, S_ISBLK(statbuf.st_mode));
+	wipe_device(DEV, device_name, force);
 
 	hdr = (struct swap_header_v1_2 *) signature_page;
 	hdr->version = 1;
@@ -659,15 +626,6 @@ main(int argc, char **argv) {
 			_("%s: unable to write signature page"),
 			device_name);
 
-	/*
-	 * A subsequent swapon() will fail if the signature
-	 * is not actually on disk. (This is a kernel bug.)
-	 */
-#ifdef HAVE_FSYNC
-	if (fsync(DEV))
-		errx(EXIT_FAILURE, _("fsync failed"));
-#endif
-
 #ifdef HAVE_LIBSELINUX
 	if (S_ISREG(statbuf.st_mode) && is_selinux_enabled() > 0) {
 		security_context_t context_string;
@@ -698,5 +656,12 @@ main(int argc, char **argv) {
 		freecon(oldcontext);
 	}
 #endif
+	/*
+	 * A subsequent swapon() will fail if the signature
+	 * is not actually on disk. (This is a kernel bug.)
+	 * The fsync() in close_fd() will take care of writing.
+	 */
+	if (close_fd(DEV) != 0)
+		err(EXIT_FAILURE, _("write failed"));
 	return EXIT_SUCCESS;
 }
