@@ -21,6 +21,110 @@
 
 #include "linux_version.h"
 #include "mountP.h"
+#include "strutils.h"
+
+/*
+ * Kernel supports only one MS_PROPAGATION flag change by one mount(2) syscall,
+ * to bypass this restriction we call mount(2) per flag. It's really not a perfect
+ * solution, but it's the same like to execute multiple mount(8) commands.
+ *
+ * We use cxt->addmounts (additional mounts) list to keep order of the requested
+ * flags changes.
+ */
+struct libmnt_addmount *mnt_new_addmount(void)
+{
+	struct libmnt_addmount *ad = calloc(1, sizeof(*ad));
+	if (!ad)
+		return NULL;
+
+	INIT_LIST_HEAD(&ad->mounts);
+	return ad;
+}
+
+void mnt_free_addmount(struct libmnt_addmount *ad)
+{
+	if (!ad)
+		return;
+	list_del(&ad->mounts);
+	free(ad);
+}
+
+static int mnt_context_append_additional_mount(struct libmnt_context *cxt,
+					       struct libmnt_addmount *ad)
+{
+	assert(cxt);
+	assert(ad);
+
+	DBG(CXT, mnt_debug_h(cxt,
+			"mount: add additional flag: 0x%08lx",
+			ad->mountflags));
+
+	list_add_tail(&ad->mounts, &cxt->addmounts);
+	return 0;
+}
+
+static int init_propagation(struct libmnt_context *cxt)
+{
+	char *name;
+	char *opts = (char *) mnt_fs_get_vfs_options(cxt->fs);
+	size_t namesz;
+	struct libmnt_optmap const *maps[1];
+
+	if (!opts)
+		return 0;
+
+	DBG(CXT, mnt_debug_h(cxt, "mount: initialize additional propagation mounts"));
+
+	maps[0] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
+
+	while (!mnt_optstr_next_option(&opts, &name, &namesz, NULL, NULL)) {
+		const struct libmnt_optmap *ent;
+		struct libmnt_addmount *ad;
+		int rc;
+
+		if (!mnt_optmap_get_entry(maps, 1, name, namesz, &ent)
+		    || !ent
+		    || !(ent->id & MS_PROPAGATION))
+			continue;
+
+		ad = mnt_new_addmount();
+		if (!ad)
+			return -ENOMEM;
+
+		ad->mountflags = ent->id;
+		rc = mnt_context_append_additional_mount(cxt, ad);
+		if (rc)
+			return rc;
+
+		cxt->mountflags &= ~ent->id;
+	}
+
+	return 0;
+}
+
+#if defined(HAVE_LIBSELINUX) || defined(HAVE_SMACK)
+struct libmnt_optname {
+	const char *name;
+	size_t namesz;
+};
+
+#define DEF_OPTNAME(n)		{ .name = n, .namesz = sizeof(n) - 1 }
+#define DEF_OPTNAME_LAST	{ .name = NULL }
+
+static int is_option(const char *name, size_t namesz,
+		     const struct libmnt_optname *names)
+{
+	const struct libmnt_optname *p;
+
+	for (p = names; p && p->name; p++) {
+		if (p->namesz == namesz
+		    && strncmp(name, p->name, namesz) == 0)
+			return 1;
+	}
+
+	return 0;
+}
+#endif /* HAVE_LIBSELINUX || HAVE_SMACK */
 
 /*
  * this has to be called after mnt_context_evaluate_permissions()
@@ -34,6 +138,25 @@ static int fix_optstr(struct libmnt_context *cxt)
 	struct libmnt_fs *fs;
 #ifdef HAVE_LIBSELINUX
 	int se_fix = 0, se_rem = 0;
+	static const struct libmnt_optname selinux_options[] = {
+		DEF_OPTNAME("context"),
+		DEF_OPTNAME("fscontext"),
+		DEF_OPTNAME("defcontext"),
+		DEF_OPTNAME("rootcontext"),
+		DEF_OPTNAME("seclabel"),
+		DEF_OPTNAME_LAST
+	};
+#endif
+#ifdef HAVE_SMACK
+	int sm_rem = 0;
+	static const struct libmnt_optname smack_options[] = {
+		DEF_OPTNAME("smackfsdef"),
+		DEF_OPTNAME("smackfsfloor"),
+		DEF_OPTNAME("smackfshat"),
+		DEF_OPTNAME("smackfsroot"),
+		DEF_OPTNAME("smackfstransmute"),
+		DEF_OPTNAME_LAST
+	};
 #endif
 	assert(cxt);
 	assert(cxt->fs);
@@ -48,15 +171,10 @@ static int fix_optstr(struct libmnt_context *cxt)
 
 	fs = cxt->fs;
 
-	/* The propagation flags should not be used together with any other
-	 * flags (except MS_REC and MS_SILENT) */
-	if (cxt->mountflags & MS_PROPAGATION)
-		cxt->mountflags &= (MS_PROPAGATION | MS_REC | MS_SILENT);
-
 	/*
 	 * The "user" options is our business (so we can modify the option),
-	 * but exception is command line for /sbin/mount.<type> helpers. Let's
-	 * save the original user=<name> to call the helpers with unchanged
+	 * the exception is command line for /sbin/mount.<type> helpers. Let's
+	 * save the original user=<name> to call the helpers with an unchanged
 	 * "user" setting.
 	 */
 	if (cxt->user_mountflags & MNT_MS_USER) {
@@ -94,6 +212,11 @@ static int fix_optstr(struct libmnt_context *cxt)
 		free(fs->user_optstr);
 		fs->user_optstr = NULL;
 	}
+	if (cxt->mountflags & MS_PROPAGATION) {
+		rc = init_propagation(cxt);
+		if (rc)
+			return rc;
+	}
 
 	next = fs->fs_optstr;
 
@@ -118,12 +241,14 @@ static int fix_optstr(struct libmnt_context *cxt)
 
 	if (!se_rem) {
 		/* de-duplicate SELinux options */
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "context");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "fscontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "defcontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "rootcontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "seclabel");
+		const struct libmnt_optname *p;
+		for (p = selinux_options; p && p->name; p++)
+			mnt_optstr_deduplicate_option(&fs->fs_optstr, p->name);
 	}
+#endif
+#ifdef HAVE_SMACK
+	if (access("/sys/fs/smackfs", F_OK) != 0)
+		sm_rem = 1;
 #endif
 	while (!mnt_optstr_next_option(&next, &name, &namesz, &val, &valsz)) {
 
@@ -132,12 +257,9 @@ static int fix_optstr(struct libmnt_context *cxt)
 		else if (namesz == 3 && !strncmp(name, "gid", 3))
 			rc = mnt_optstr_fix_gid(&fs->fs_optstr, val, valsz, &next);
 #ifdef HAVE_LIBSELINUX
-		else if ((se_rem || se_fix) &&
-			 namesz >= 7 && (!strncmp(name, "context", 7) ||
-					 !strncmp(name, "fscontext", 9) ||
-					 !strncmp(name, "defcontext", 10) ||
-					 !strncmp(name, "rootcontext", 11) ||
-					 !strncmp(name, "seclabel", 8))) {
+		else if ((se_rem || se_fix)
+			 && is_option(name, namesz, selinux_options)) {
+
 			if (se_rem) {
 				/* remove context= option */
 				next = name;
@@ -150,6 +272,15 @@ static int fix_optstr(struct libmnt_context *cxt)
 				rc = mnt_optstr_fix_secontext(&fs->fs_optstr,
 							val, valsz, &next);
 		}
+#endif
+#ifdef HAVE_SMACK
+		else if (sm_rem && is_option(name, namesz, smack_options)) {
+
+			next = name;
+			rc = mnt_optstr_remove_option_at(&fs->fs_optstr,
+					name,
+					val ? val + valsz : name + namesz);
+		}
 #endif
 		if (rc)
 			goto done;
@@ -175,12 +306,12 @@ done:
 }
 
 /*
- * Converts already evaluated and fixed options to the form that is compatible
+ * Converts the already evaluated and fixed options to the form that is compatible
  * with /sbin/mount.type helpers.
  */
 static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 {
-	struct libmnt_optmap const *maps[1];
+	struct libmnt_optmap const *maps[2];
 	char *next, *name, *val;
 	size_t namesz, valsz;
 	int rc = 0;
@@ -195,6 +326,28 @@ static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 	if (!*optstr)
 		return -ENOMEM;
 
+	if (cxt->user_mountflags & MNT_MS_USER) {
+		/*
+		 * This is unnecessary for real user-mounts as mount.<type>
+		 * helpers always have to follow fstab rather than mount
+		 * options on the command line.
+		 *
+		 * However, if you call mount.<type> as root, then the helper follows
+		 * the command line. If there is (for example) "user,exec" in fstab,
+		 * then we have to manually append the "exec" back to the options
+		 * string, bacause there is nothing like MS_EXEC (we only have
+		 * MS_NOEXEC in mount flags and we don't care about the original
+		 * mount string in libmount for VFS options).
+		 */
+		if (!(cxt->mountflags & MS_NOEXEC))
+			mnt_optstr_append_option(optstr, "exec", NULL);
+		if (!(cxt->mountflags & MS_NOSUID))
+			mnt_optstr_append_option(optstr, "suid", NULL);
+		if (!(cxt->mountflags & MS_NODEV))
+			mnt_optstr_append_option(optstr, "dev", NULL);
+	}
+
+
 	if (cxt->flags & MNT_FL_SAVED_USER)
 		rc = mnt_optstr_set_option(optstr, "user", cxt->orig_user);
 	if (rc)
@@ -202,12 +355,13 @@ static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 
 	/* remove userspace options with MNT_NOHLPS flag */
 	maps[0] = mnt_get_builtin_optmap(MNT_USERSPACE_MAP);
+	maps[1] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
 	next = *optstr;
 
 	while (!mnt_optstr_next_option(&next, &name, &namesz, &val, &valsz)) {
 		const struct libmnt_optmap *ent;
 
-		mnt_optmap_get_entry(maps, 1, name, namesz, &ent);
+		mnt_optmap_get_entry(maps, 2, name, namesz, &ent);
 		if (ent && ent->id && (ent->mask & MNT_NOHLPS)) {
 			next = name;
 			rc = mnt_optstr_remove_option_at(optstr, name,
@@ -227,7 +381,7 @@ err:
 /*
  * this has to be called before fix_optstr()
  *
- * Note that user=<name> maybe be used by some filesystems as filesystem
+ * Note that user=<name> may be used by some filesystems as a filesystem
  * specific option (e.g. cifs). Yes, developers of such filesystems have
  * allocated pretty hot place in hell...
  */
@@ -258,7 +412,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 		/*
 		 * user mount
 		 */
-		if (!(cxt->flags & MNT_FL_TAB_APPLIED))
+		if (!mnt_context_tab_applied(cxt))
 		{
 			DBG(CXT, mnt_debug_h(cxt, "perms: fstab not applied, ignore user mount"));
 			return -EPERM;
@@ -330,7 +484,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 /*
  * mnt_context_helper_setopt() backend
  *
- * This function applies mount.type command line option (for example parsed
+ * This function applies the mount.type command line option (for example parsed
  * by getopt() or getopt_long()) to @cxt. All unknown options are ignored and
  * then 1 is returned.
  *
@@ -416,7 +570,7 @@ static int exec_helper(struct libmnt_context *cxt)
 
 		/*
 		 * TODO: remove the exception for "nfs", -s is documented
-		 *       for years should be usable everywhere.
+		 *       for years and should be usable everywhere.
 		 */
 		if (mnt_context_is_sloppy(cxt) &&
 		    type && startswith(type, "nfs"))
@@ -467,9 +621,46 @@ static int exec_helper(struct libmnt_context *cxt)
 	return rc;
 }
 
+static int do_mount_additional(struct libmnt_context *cxt,
+			       const char *target,
+			       unsigned long flags,
+			       int *syserr)
+{
+	struct list_head *p;
+
+	assert(cxt);
+	assert(target);
+
+	if (syserr)
+		*syserr = 0;
+
+	list_for_each(p, &cxt->addmounts) {
+		int rc;
+		struct libmnt_addmount *ad =
+				list_entry(p, struct libmnt_addmount, mounts);
+
+		DBG(CXT, mnt_debug_h(cxt, "mount(2) changing flag: 0x%08lx %s",
+				ad->mountflags,
+				ad->mountflags & MS_REC ? " (recursive)" : ""));
+
+		rc = mount("none", target, NULL,
+				ad->mountflags | (flags & MS_SILENT), NULL);
+		if (rc) {
+			if (syserr)
+				*syserr = -errno;
+			DBG(CXT, mnt_debug_h(cxt,
+					"mount(2) failed [errno=%d %m]",
+					errno));
+			return rc;
+		}
+	}
+
+	return 0;
+}
+
 /*
  * The default is to use fstype from cxt->fs, this could be overwritten by
- * @try_type argument.
+ * @try_type argument. If @try_type is specified then mount with MS_SILENT.
  *
  * Returns: 0 on success,
  *         >0 in case of mount(2) error (returns syscall errno),
@@ -490,18 +681,28 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 		if (rc)
 			return rc;
 	}
-	if (cxt->helper)
-		return exec_helper(cxt);
 
 	flags = cxt->mountflags;
 	src = mnt_fs_get_srcpath(cxt->fs);
 	target = mnt_fs_get_target(cxt->fs);
 
+	if (cxt->helper) {
+		rc = exec_helper(cxt);
+
+		if (mnt_context_helper_executed(cxt)
+		    && mnt_context_get_helper_status(cxt) == 0
+		    && !list_empty(&cxt->addmounts)
+		    && do_mount_additional(cxt, target, flags, NULL))
+
+			return -MNT_ERR_APPLYFLAGS;
+		return rc;
+	}
+
 	if (!target)
 		return -EINVAL;
 	if (!src) {
 		/* unnecessary, should be already resolved in
-		 * mnt_context_prepare_srcpath(), but for sure... */
+		 * mnt_context_prepare_srcpath(), but to be sure... */
 		DBG(CXT, mnt_debug_h(cxt, "WARNING: source is NULL -- using \"none\"!"));
 		src = "none";
 	}
@@ -509,6 +710,8 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 
 	if (!(flags & MS_MGC_MSK))
 		flags |= MS_MGC_VAL;
+	if (try_type)
+		flags |= MS_SILENT;
 
 	DBG(CXT, mnt_debug_h(cxt, "%smount(2) "
 			"[source=%s, target=%s, type=%s, "
@@ -517,9 +720,22 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 			src, target, type,
 			flags, cxt->mountdata ? "yes" : "<none>"));
 
-	if (mnt_context_is_fake(cxt))
+	if (mnt_context_is_fake(cxt)) {
+		/*
+		 * fake
+		 */
 		cxt->syscall_status = 0;
-	else {
+
+	} else if (mnt_context_propagation_only(cxt)) {
+		/*
+		 * propagation flags *only*
+		 */
+		if (do_mount_additional(cxt, target, flags, &cxt->syscall_status))
+			return -MNT_ERR_APPLYFLAGS;
+	} else {
+		/*
+		 * regular mount
+		 */
 		if (mount(src, target, type, flags, cxt->mountdata)) {
 			cxt->syscall_status = -errno;
 			DBG(CXT, mnt_debug_h(cxt, "mount(2) failed [errno=%d %m]",
@@ -528,6 +744,16 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 		}
 		DBG(CXT, mnt_debug_h(cxt, "mount(2) success"));
 		cxt->syscall_status = 0;
+
+		/*
+		 * additional mounts for extra propagation flags
+		 */
+		if (!list_empty(&cxt->addmounts)
+		    && do_mount_additional(cxt, target, flags, NULL)) {
+
+			/* TODO: call umount? */
+			return -MNT_ERR_APPLYFLAGS;
+		}
 	}
 
 	if (try_type && cxt->update) {
@@ -626,7 +852,6 @@ int mnt_context_prepare_mount(struct libmnt_context *cxt)
 
 	DBG(CXT, mnt_debug_h(cxt, "mount: preparing"));
 
-	/* TODO: fstab is unnecessary for MS_{MOVE,BIND,STARED,...} */
 	rc = mnt_context_apply_fstab(cxt);
 	if (!rc)
 		rc = mnt_context_merge_mflags(cxt);
@@ -657,10 +882,10 @@ int mnt_context_prepare_mount(struct libmnt_context *cxt)
  * Call mount(2) or mount.type helper. Unnecessary for mnt_context_mount().
  *
  * Note that this function could be called only once. If you want to mount
- * another source or target than you have to call mnt_reset_context().
+ * another source or target, then you have to call mnt_reset_context().
  *
- * If you want to call mount(2) for the same source and target with a different
- * mount flags or fstype then call mnt_context_reset_status() and then try
+ * If you want to call mount(2) for the same source and target with different
+ * mount flags or fstype, then call mnt_context_reset_status() and then try
  * again mnt_context_do_mount().
  *
  * WARNING: non-zero return code does not mean that mount(2) syscall or
@@ -693,7 +918,7 @@ int mnt_context_do_mount(struct libmnt_context *cxt)
 	type = mnt_fs_get_fstype(cxt->fs);
 	if (type) {
 		if (strchr(type, ','))
-			/* this only happen if fstab countains list of filesystems */
+			/* this only happens if fstab contains a list of filesystems */
 			res = do_mount_by_pattern(cxt, type);
 		else
 			res = do_mount(cxt, NULL);
@@ -722,7 +947,8 @@ int mnt_context_do_mount(struct libmnt_context *cxt)
 		 * system that does not have write support. Check this to avoid
 		 * 'ro' in /proc/mounts and 'rw' in mtab.
 		 */
-		if (!(cxt->mountflags & (MS_RDONLY | MS_PROPAGATION | MS_MOVE))
+		if (!(cxt->mountflags & (MS_RDONLY | MS_MOVE))
+		    && !mnt_context_propagation_only(cxt)
 		    && mnt_is_readonly(mnt_context_get_target(cxt)))
 
 			mnt_context_set_mflags(cxt,
@@ -760,7 +986,7 @@ int mnt_context_finalize_mount(struct libmnt_context *cxt)
  * mnt_context_mount:
  * @cxt: mount context
  *
- * High-level, mounts filesystem by mount(2) or fork()+exec(/sbin/mount.type).
+ * High-level, mounts the filesystem by mount(2) or fork()+exec(/sbin/mount.type).
  *
  * This is similar to:
  *
@@ -770,10 +996,10 @@ int mnt_context_finalize_mount(struct libmnt_context *cxt)
  *
  * See also mnt_context_disable_helpers().
  *
- * Note that this function could be called only once. If you want to mount with
- * different setting than you have to call mnt_reset_context(). It's NOT enough
- * to call mnt_context_reset_status() if you want call this function more than
- * once, whole context has to be reseted.
+ * Note that this function should be called only once. If you want to mount with
+ * different settings, then you have to call mnt_reset_context(). It's NOT enough
+ * to call mnt_context_reset_status(). If you want to call this function more than
+ * once, the whole context has to be reset.
  *
  * WARNING: non-zero return code does not mean that mount(2) syscall or
  *          mount.type helper wasn't successfully called.
@@ -798,11 +1024,6 @@ int mnt_context_mount(struct libmnt_context *cxt)
 		rc = mnt_context_prepare_update(cxt);
 	if (!rc)
 		rc = mnt_context_do_mount(cxt);
-
-	/* TODO: if mtab update is expected then check if the
-	 * target is really mounted read-write to avoid 'ro' in
-	 * mtab and 'rw' in /proc/mounts.
-	 */
 	if (!rc)
 		rc = mnt_context_update_tabs(cxt);
 	return rc;
@@ -814,7 +1035,7 @@ int mnt_context_mount(struct libmnt_context *cxt)
  * @itr: iterator
  * @fs: returns the current filesystem
  * @mntrc: returns the return code from mnt_context_mount()
- * @ignored: returns 1 for not matching and 2 for already mounted filesystems
+ * @ignored: returns 1 for non-matching and 2 for already mounted filesystems
  *
  * This function tries to mount the next filesystem from fstab (as returned by
  * mnt_context_get_fstab()). See also mnt_context_set_fstab().
@@ -882,11 +1103,11 @@ int mnt_context_next_mount(struct libmnt_context *cxt,
 	/* ignore noauto filesystems */
 	   (o && mnt_optstr_get_option(o, "noauto", NULL, NULL) == 0) ||
 
-	/* ignore filesystems not match with options patterns */
+	/* ignore filesystems which don't match options patterns */
 	   (cxt->fstype_pattern && !mnt_fs_match_fstype(*fs,
 					cxt->fstype_pattern)) ||
 
-	/* ignore filesystems not match with type patterns */
+	/* ignore filesystems which don't match type patterns */
 	   (cxt->optstr_pattern && !mnt_fs_match_options(*fs,
 					cxt->optstr_pattern))) {
 		if (ignored)
