@@ -133,7 +133,7 @@ static struct colinfo infos[FINDMNT_NCOLUMNS] = {
 	[COL_ID]           = { "ID",              2, TT_FL_RIGHT, N_("mount ID") },
 	[COL_OPT_FIELDS]   = { "OPT-FIELDS",   0.10, TT_FL_TRUNC, N_("optional mount fields") },
 	[COL_PROPAGATION]  = { "PROPAGATION",  0.10, 0, N_("VFS propagation flags") },
-	[COL_FREQ]         = { "FREQ",            1, TT_FL_RIGHT, N_("dump(8) frequency in days [fstab only]") },
+	[COL_FREQ]         = { "FREQ",            1, TT_FL_RIGHT, N_("dump(8) period in days [fstab only]") },
 	[COL_PASSNO]       = { "PASSNO",          1, TT_FL_RIGHT, N_("pass number on parallel fsck(8) [fstab only]") }
 };
 
@@ -368,17 +368,24 @@ static char *get_tag_from_udev(const char *devname, int col)
 {
 	struct udev_device *dev;
 	const char *data = NULL;
-	char *res = NULL;
+	char *res = NULL, *path;
 
 	if (!udev)
 		udev = udev_new();
 	if (!udev)
 		return NULL;
 
+	/* libudev don't like /dev/mapper/ symlinks */
+	path = realpath(devname, NULL);
+	if (path)
+		devname = path;
+
 	if (strncmp(devname, "/dev/", 5) == 0)
 		devname += 5;
 
 	dev = udev_device_new_from_subsystem_sysname(udev, "block", devname);
+	free(path);
+
 	if (!dev)
 		return NULL;
 
@@ -410,12 +417,17 @@ static char *get_tag_from_udev(const char *devname, int col)
 #endif /* HAVE_LIBUDEV */
 
 /* Returns LABEL or UUID */
-static const char *get_tag(struct libmnt_fs *fs, const char *tagname, int col)
+static char *get_tag(struct libmnt_fs *fs, const char *tagname, int col
+#ifndef HAVE_LIBUDEV
+		__attribute__((__unused__))
+#endif
+		)
 {
-	const char *t, *v, *res = NULL;
+	const char *t, *v;
+	char *res = NULL;
 
 	if (!mnt_fs_get_tag(fs, &t, &v) && !strcmp(t, tagname))
-		res = v;
+		res = xstrdup(v);
 	else {
 		const char *dev = mnt_fs_get_source(fs);
 
@@ -425,14 +437,18 @@ static const char *get_tag(struct libmnt_fs *fs, const char *tagname, int col)
 		if (dev)
 			res = get_tag_from_udev(dev, col);
 #endif
-		if (!res)
+		if (!res) {
 			res = mnt_cache_find_tag_value(cache, dev, tagname);
+			if (res && cache)
+				/* don't return pointer to cache */
+				res = xstrdup(res);
+		}
 	}
 
 	return res;
 }
 
-static const char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
+static char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
 {
 	struct statvfs buf;
 	uint64_t vfs_attr = 0;
@@ -453,7 +469,7 @@ static const char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
 		break;
 	case COL_USEPERC:
 		if (buf.f_blocks == 0)
-			return "-";
+			return xstrdup("-");
 
 		xasprintf(&sizestr, "%.0f%%",
 				(double)(buf.f_blocks - buf.f_bfree) /
@@ -461,56 +477,54 @@ static const char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
 		return sizestr;
 	}
 
-	return vfs_attr == 0 ? "0" :
+	return vfs_attr == 0 ? xstrdup("0") :
 		size_to_human_string(SIZE_SUFFIX_1LETTER, vfs_attr);
 }
 
 /* reads FS data from libmount
  */
-static const char *get_data(struct libmnt_fs *fs, int num)
+static char *get_data(struct libmnt_fs *fs, int num)
 {
-	const char *str = NULL;
-	char *tmp;
+	char *str = NULL;
 	int col_id = get_column_id(num);
 
 	switch (col_id) {
 	case COL_SOURCE:
 	{
 		const char *root = mnt_fs_get_root(fs);
+		const char *spec = mnt_fs_get_srcpath(fs);
 
-		str = mnt_fs_get_srcpath(fs);
+		if (spec && (flags & FL_CANONICALIZE))
+			spec = mnt_resolve_path(spec, cache);
+		if (!spec) {
+			spec = mnt_fs_get_source(fs);
 
-		if (str && (flags & FL_CANONICALIZE))
-			str = mnt_resolve_path(str, cache);
-		if (!str) {
-			str = mnt_fs_get_source(fs);
-
-			if (str && (flags & FL_EVALUATE))
-				str = mnt_resolve_spec(str, cache);
-		}
-		if (root && str && !(flags & FL_NOFSROOT) && strcmp(root, "/")) {
-			xasprintf(&tmp, "%s[%s]", str, root);
-			str = tmp;
+			if (spec && (flags & FL_EVALUATE))
+				spec = mnt_resolve_spec(spec, cache);
 		}
+		if (root && spec && !(flags & FL_NOFSROOT) && strcmp(root, "/"))
+			xasprintf(&str, "%s[%s]", spec, root);
+		else if (spec)
+			str = xstrdup(spec);
 		break;
 	}
 	case COL_TARGET:
-		str = mnt_fs_get_target(fs);
+		str = xstrdup(mnt_fs_get_target(fs));
 		break;
 	case COL_FSTYPE:
-		str = mnt_fs_get_fstype(fs);
+		str = xstrdup(mnt_fs_get_fstype(fs));
 		break;
 	case COL_OPTIONS:
-		str = mnt_fs_get_options(fs);
+		str = xstrdup(mnt_fs_get_options(fs));
 		break;
 	case COL_VFS_OPTIONS:
-		str = mnt_fs_get_vfs_options(fs);
+		str = xstrdup(mnt_fs_get_vfs_options(fs));
 		break;
 	case COL_FS_OPTIONS:
-		str = mnt_fs_get_fs_options(fs);
+		str = xstrdup(mnt_fs_get_fs_options(fs));
 		break;
 	case COL_OPT_FIELDS:
-		str = mnt_fs_get_optional_fields(fs);
+		str = xstrdup(mnt_fs_get_optional_fields(fs));
 		break;
 	case COL_UUID:
 		str = get_tag(fs, "UUID", col_id);
@@ -532,10 +546,9 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 			break;
 
 		if ((tt_flags & TT_FL_RAW) || (tt_flags & TT_FL_EXPORT))
-			xasprintf(&tmp, "%u:%u", major(devno), minor(devno));
+			xasprintf(&str, "%u:%u", major(devno), minor(devno));
 		else
-			xasprintf(&tmp, "%3u:%-3u", major(devno), minor(devno));
-		str = tmp;
+			xasprintf(&str, "%3u:%-3u", major(devno), minor(devno));
 		break;
 	}
 	case COL_SIZE:
@@ -545,19 +558,15 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 		str = get_vfs_attr(fs, col_id);
 		break;
 	case COL_FSROOT:
-		str = mnt_fs_get_root(fs);
+		str = xstrdup(mnt_fs_get_root(fs));
 		break;
 	case COL_TID:
-		if (mnt_fs_get_tid(fs)) {
-			xasprintf(&tmp, "%d", mnt_fs_get_tid(fs));
-			str = tmp;
-		}
+		if (mnt_fs_get_tid(fs))
+			xasprintf(&str, "%d", mnt_fs_get_tid(fs));
 		break;
 	case COL_ID:
-		if (mnt_fs_get_id(fs)) {
-			xasprintf(&tmp, "%d", mnt_fs_get_id(fs));
-			str = tmp;
-		}
+		if (mnt_fs_get_id(fs))
+			xasprintf(&str, "%d", mnt_fs_get_id(fs));
 		break;
 	case COL_PROPAGATION:
 		if (mnt_fs_is_kernel(fs)) {
@@ -570,29 +579,25 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 			n = xstrdup((fl & MS_SHARED) ? "shared" : "private");
 
 			if (fl & MS_SLAVE) {
-				xasprintf(&tmp, "%s,slave", n);
+				xasprintf(&str, "%s,slave", n);
 				free(n);
-				n = tmp;
+				n = str;
 			}
 			if (fl & MS_UNBINDABLE) {
-				xasprintf(&tmp, "%s,unbindable", n);
+				xasprintf(&str, "%s,unbindable", n);
 				free(n);
-				n = tmp;
+				n = str;
 			}
 			str = n;
 		}
 		break;
 	case COL_FREQ:
-		if (!mnt_fs_is_kernel(fs)) {
-			xasprintf(&tmp, "%d", mnt_fs_get_freq(fs));
-			str = tmp;
-		}
+		if (!mnt_fs_is_kernel(fs))
+			xasprintf(&str, "%d", mnt_fs_get_freq(fs));
 		break;
 	case COL_PASSNO:
-		if (!mnt_fs_is_kernel(fs)) {
-			xasprintf(&tmp, "%d", mnt_fs_get_passno(fs));
-			str = tmp;
-		}
+		if (!mnt_fs_is_kernel(fs))
+			xasprintf(&str, "%d", mnt_fs_get_passno(fs));
 		break;
 	default:
 		break;
@@ -600,12 +605,12 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 	return str;
 }
 
-static const char *get_tabdiff_data(struct libmnt_fs *old_fs,
+static char *get_tabdiff_data(struct libmnt_fs *old_fs,
 				    struct libmnt_fs *new_fs,
 				    int change,
 				    int num)
 {
-	const char *str = NULL;
+	char *str = NULL;
 
 	switch (get_column_id(num)) {
 	case COL_ACTION:
@@ -626,16 +631,17 @@ static const char *get_tabdiff_data(struct libmnt_fs *old_fs,
 			str = _("unknown");
 			break;
 		}
+		str = xstrdup(str);
 		break;
 	case COL_OLD_OPTIONS:
 		if (old_fs && (change == MNT_TABDIFF_REMOUNT ||
 			       change == MNT_TABDIFF_UMOUNT))
-			str = mnt_fs_get_options(old_fs);
+			str = xstrdup(mnt_fs_get_options(old_fs));
 		break;
 	case COL_OLD_TARGET:
 		if (old_fs && (change == MNT_TABDIFF_MOVE ||
 			       change == MNT_TABDIFF_UMOUNT))
-			str = mnt_fs_get_target(old_fs);
+			str = xstrdup(mnt_fs_get_target(old_fs));
 		break;
 	default:
 		if (new_fs)
@@ -795,7 +801,7 @@ static struct libmnt_table *parse_tabfiles(char **files,
 			break;
 		}
 		if (rc) {
-			mnt_free_table(tb);
+			mnt_unref_table(tb);
 			warn(_("can't read %s"), path);
 			return NULL;
 		}
@@ -815,8 +821,9 @@ static int tab_is_tree(struct libmnt_table *tb)
 	if (!itr)
 		return 0;
 
-	if (mnt_table_next_fs(tb, itr, &fs) == 0)
-		rc = mnt_fs_get_id(fs) > 0 && mnt_fs_get_parent_id(fs) > 0;
+	rc = (mnt_table_next_fs(tb, itr, &fs) == 0 &&
+	      mnt_fs_is_kernel(fs) &&
+	      mnt_fs_get_root(fs));
 
 	mnt_free_iter(itr);
 	return rc;
@@ -831,14 +838,6 @@ static int match_func(struct libmnt_fs *fs,
 	const char *m;
 	void *md;
 
-	m = get_match(COL_TARGET);
-	if (m && !mnt_fs_match_target(fs, m, cache))
-		return rc;
-
-	m = get_match(COL_SOURCE);
-	if (m && !mnt_fs_match_source(fs, m, cache))
-		return rc;
-
 	m = get_match(COL_FSTYPE);
 	if (m && !mnt_fs_match_fstype(fs, m))
 		return rc;
@@ -851,6 +850,14 @@ static int match_func(struct libmnt_fs *fs,
 	if (md && mnt_fs_get_devno(fs) != *((dev_t *) md))
 		return rc;
 
+	m = get_match(COL_TARGET);
+	if (m && !mnt_fs_match_target(fs, m, cache))
+		return rc;
+
+	m = get_match(COL_SOURCE);
+	if (m && !mnt_fs_match_source(fs, m, cache))
+		return rc;
+
 	if ((flags & FL_DF) && !(flags & FL_ALL)) {
 		const char *type = mnt_fs_get_fstype(fs);
 
@@ -1076,7 +1083,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 
 	rc = 0;
 done:
-	mnt_free_table(tb_new);
+	mnt_unref_table(tb_new);
 	mnt_free_tabdiff(diff);
 	mnt_free_iter(itr);
 	if (f)
@@ -1424,7 +1431,7 @@ int main(int argc, char *argv[])
 	/*
 	 * initialize output formatting (tt.h)
 	 */
-	tt = tt_new_table(tt_flags);
+	tt = tt_new_table(tt_flags | TT_FL_FREEDATA);
 	if (!tt) {
 		warn(_("failed to initialize output table"));
 		goto leave;
@@ -1485,8 +1492,9 @@ int main(int argc, char *argv[])
 leave:
 	tt_free_table(tt);
 
-	mnt_free_table(tb);
-	mnt_free_cache(cache);
+	mnt_unref_table(tb);
+	mnt_unref_cache(cache);
+
 	free(tabfiles);
 #ifdef HAVE_LIBUDEV
 	udev_unref(udev);
