@@ -115,7 +115,7 @@ static const struct dmesg_name facility_names[] =
 	[FAC_BASE(LOG_UUCP)]     = { "uucp",     N_("UUCP subsystem") },
 	[FAC_BASE(LOG_CRON)]     = { "cron",     N_("clock daemon") },
 	[FAC_BASE(LOG_AUTHPRIV)] = { "authpriv", N_("security/authorization messages (private)") },
-	[FAC_BASE(LOG_FTP)]      = { "ftp",      N_("ftp daemon") },
+	[FAC_BASE(LOG_FTP)]      = { "ftp",      N_("FTP daemon") },
 };
 
 /* supported methods to read message buffer
@@ -126,6 +126,18 @@ enum {
 	DMESG_METHOD_MMAP	/* mmap file with records (see --file) */
 };
 
+enum {
+	DMESG_TIMEFTM_NONE = 0,
+	DMESG_TIMEFTM_CTIME,		/* [ctime] */
+	DMESG_TIMEFTM_CTIME_DELTA,	/* [ctime <delta>] */
+	DMESG_TIMEFTM_DELTA,		/* [<delta>] */
+	DMESG_TIMEFTM_RELTIME,		/* [relative] */
+	DMESG_TIMEFTM_TIME,		/* [time] */
+	DMESG_TIMEFTM_TIME_DELTA,	/* [time <delta>] */
+	DMESG_TIMEFTM_ISO8601		/* 2013-06-13T22:11:00,123456+0100 */
+};
+#define is_timefmt(c, f) ((c)->time_fmt == (DMESG_TIMEFTM_ ##f))
+
 struct dmesg_control {
 	/* bit arrays -- see include/bitops.h */
 	char levels[ARRAY_SIZE(level_names) / NBBY + 1];
@@ -133,7 +145,7 @@ struct dmesg_control {
 
 	struct timeval	lasttime;	/* last printed timestamp */
 	struct tm	lasttm;		/* last localtime */
-	time_t		boot_time;	/* system boot time */
+	struct timeval	boot_time;	/* system boot time */
 
 	int		action;		/* SYSLOG_ACTION_* */
 	int		method;		/* DMESG_METHOD_* */
@@ -152,16 +164,13 @@ struct dmesg_control {
 	char		*filename;
 	char		*mmap_buff;
 	size_t		pagesize;
+	unsigned int	time_fmt;	/* time format */
 
 	unsigned int	follow:1,	/* wait for new messages */
 			raw:1,		/* raw mode */
 			fltr_lev:1,	/* filter out by levels[] */
 			fltr_fac:1,	/* filter out by facilities[] */
 			decode:1,	/* use "facility: level: " prefix */
-			notime:1,	/* don't print timestamp */
-			delta:1,	/* show time deltas */
-			reltime:1,	/* show human readable relative times */
-			ctime:1,	/* show human readable time */
 			pager:1,	/* pipe output into a pager */
 			color:1;	/* colorize messages */
 };
@@ -229,26 +238,28 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	fputs(_(" -C, --clear                 clear the kernel ring buffer\n"), out);
 	fputs(_(" -c, --read-clear            read and clear all messages\n"), out);
 	fputs(_(" -D, --console-off           disable printing messages to console\n"), out);
-	fputs(_(" -d, --show-delta            show time delta between printed messages\n"), out);
-	fputs(_(" -e, --reltime               show local time and time delta in readable format\n"), out);
 	fputs(_(" -E, --console-on            enable printing messages to console\n"), out);
 	fputs(_(" -F, --file <file>           use the file instead of the kernel log buffer\n"), out);
 	fputs(_(" -f, --facility <list>       restrict output to defined facilities\n"), out);
 	fputs(_(" -H, --human                 human readable output\n"), out);
 	fputs(_(" -k, --kernel                display kernel messages\n"), out);
-	fputs(_(" -L, --color                 colorize messages\n"), out);
+	fputs(_(" -L, --color[=<when>]        colorize messages (auto, always or never)\n"), out);
 	fputs(_(" -l, --level <list>          restrict output to defined levels\n"), out);
 	fputs(_(" -n, --console-level <level> set level of messages printed to console\n"), out);
 	fputs(_(" -P, --nopager               do not pipe output into a pager\n"), out);
 	fputs(_(" -r, --raw                   print the raw message buffer\n"), out);
 	fputs(_(" -S, --syslog                force to use syslog(2) rather than /dev/kmsg\n"), out);
 	fputs(_(" -s, --buffer-size <size>    buffer size to query the kernel ring buffer\n"), out);
-	fputs(_(" -T, --ctime                 show human readable timestamp (could be \n"
-		"                               inaccurate if you have used SUSPEND/RESUME)\n"), out);
-	fputs(_(" -t, --notime                don't print messages timestamp\n"), out);
 	fputs(_(" -u, --userspace             display userspace messages\n"), out);
 	fputs(_(" -w, --follow                wait for new messages\n"), out);
 	fputs(_(" -x, --decode                decode facility and level to readable string\n"), out);
+	fputs(_(" -d, --show-delta            show time delta between printed messages\n"), out);
+	fputs(_(" -e, --reltime               show local time and time delta in readable format\n"), out);
+	fputs(_(" -T, --ctime                 show human readable timestamp\n"), out);
+	fputs(_(" -t, --notime                don't print messages timestamp\n"), out);
+	fputs(_("     --time-format <format>  show time stamp using format:\n"
+		"                               [delta|reltime|ctime|notime|iso]\n"
+		"Suspending/resume will make ctime and iso timestamps inaccurate.\n"), out);
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
 	fputs(USAGE_VERSION, out);
@@ -264,7 +275,7 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 			level_names[i].name,
 			_(level_names[i].help));
 	fputs(USAGE_SEPARATOR, out);
-	fprintf(out, USAGE_MAN_TAIL("dmesg(q)"));
+	fprintf(out, USAGE_MAN_TAIL("dmesg(1)"));
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
@@ -461,17 +472,30 @@ static int get_syslog_buffer_size(void)
 	return n > 0 ? n : 0;
 }
 
-static time_t get_boot_time(void)
+static int get_boot_time(struct timeval *boot_time)
 {
+	struct timespec hires_uptime;
+	struct timeval lores_uptime, now;
 	struct sysinfo info;
-	struct timeval tv;
 
+	if (gettimeofday(&now, NULL) != 0) {
+		warn(_("gettimeofday failed"));
+		return -errno;
+	}
+
+#ifdef CLOCK_BOOTTIME
+	if (clock_gettime(CLOCK_BOOTTIME, &hires_uptime) == 0) {
+		TIMESPEC_TO_TIMEVAL(&lores_uptime, &hires_uptime);
+		timersub(&now, &lores_uptime, boot_time);
+		return 0;
+	}
+#endif
+	/* fallback */
 	if (sysinfo(&info) != 0)
 		warn(_("sysinfo failed"));
-	else if (gettimeofday(&tv, NULL) != 0)
-		warn(_("gettimeofday failed"));
-	else
-		return tv.tv_sec -= info.uptime;
+
+	boot_time->tv_sec = now.tv_sec - info.uptime;
+	boot_time->tv_usec = 0;
 	return 0;
 }
 
@@ -615,8 +639,11 @@ static void safe_fwrite(const char *buf, size_t size, FILE *out)
 			rc = fwrite_hex(p, len, out);
 		else
 			rc = fwrite(p, 1, len, out) != len;
-		if (rc != 0)
-			err(EXIT_FAILURE, _("write failed"));
+		if (rc != 0) {
+			if (errno != EPIPE)
+				err(EXIT_FAILURE, _("write failed"));
+			exit(EXIT_SUCCESS);
+		}
 	}
 }
 
@@ -689,9 +716,10 @@ static int get_next_syslog_record(struct dmesg_control *ctl,
 
 		if (*begin == '[' && (*(begin + 1) == ' ' ||
 				      isdigit(*(begin + 1)))) {
-			if (ctl->delta || ctl->ctime || ctl->reltime)
+
+			if (!is_timefmt(ctl, NONE))
 				begin = parse_syslog_timestamp(begin + 1, &rec->tv);
-			else if (ctl->notime)
+			else
 				begin = skip_item(begin, end, "]");
 
 			if (begin < end && *begin == ' ')
@@ -759,7 +787,7 @@ static struct tm *record_localtime(struct dmesg_control *ctl,
 				   struct dmesg_record *rec,
 				   struct tm *tm)
 {
-	time_t t = ctl->boot_time + rec->tv.tv_sec;
+	time_t t = ctl->boot_time.tv_sec + rec->tv.tv_sec;
 	return localtime_r(&t, tm);
 }
 
@@ -783,6 +811,23 @@ static char *short_ctime(struct tm *tm, char *buf, size_t bufsiz)
 	return buf;
 }
 
+static char *iso_8601_time(struct dmesg_control *ctl, struct dmesg_record *rec,
+			   char *buf, size_t bufsiz)
+{
+	struct tm tm;
+	size_t len;
+	record_localtime(ctl, rec, &tm);
+	if (strftime(buf, bufsiz, "%Y-%m-%dT%H:%M:%S", &tm) == 0) {
+		*buf = '\0';
+		return buf;
+	}
+	len = strlen(buf);
+	snprintf(buf + len, bufsiz - len, ",%06d", (int)rec->tv.tv_usec);
+	len = strlen(buf);
+	strftime(buf + len, bufsiz - len, "%z", &tm);
+	return buf;
+}
+
 static double record_count_delta(struct dmesg_control *ctl,
 				 struct dmesg_record *rec)
 {
@@ -852,79 +897,56 @@ static void print_record(struct dmesg_control *ctl,
 		printf("%-6s:%-6s: ", facility_names[rec->facility].name,
 				      level_names[rec->level].name);
 
-	/*
-	 * [sec.usec <delta>] or [ctime <delta>]
-	 */
-	if (ctl->delta) {
-		if (ctl->color)
-			color_enable(DMESG_COLOR_TIME);
-		if (ctl->ctime)
-			printf("[%s ", record_ctime(ctl, rec, buf, sizeof(buf)));
-		else if (ctl->notime)
-			putchar('[');
-		else
-			printf("[%5d.%06d ", (int) rec->tv.tv_sec,
-					     (int) rec->tv.tv_usec);
-		printf("<%12.06f>] ", record_count_delta(ctl, rec));
-		if (ctl->color)
-			color_disable();
+	if (ctl->color)
+		color_enable(DMESG_COLOR_TIME);
 
-	/*
-	 * [ctime]
-	 */
-	} else if (ctl->ctime) {
-		if (ctl->color)
-			color_enable(DMESG_COLOR_TIME);
-		printf("[%s] ", record_ctime(ctl, rec, buf, sizeof(buf)));
-		if (ctl->color)
-			color_disable();
-	}
-
-	/*
-	 * [reltime]
-	 */
-	else if (ctl->reltime) {
+	switch (ctl->time_fmt) {
 		double delta;
 		struct tm cur;
-
+	case DMESG_TIMEFTM_NONE:
+		break;
+	case DMESG_TIMEFTM_CTIME:
+		printf("[%s] ", record_ctime(ctl, rec, buf, sizeof(buf)));
+		break;
+	case DMESG_TIMEFTM_CTIME_DELTA:
+		printf("[%s <%12.06f>] ",
+		       record_ctime(ctl, rec, buf, sizeof(buf)),
+		       record_count_delta(ctl, rec));
+		break;
+	case DMESG_TIMEFTM_DELTA:
+		printf("[<%12.06f>] ", record_count_delta(ctl, rec));
+		break;
+	case DMESG_TIMEFTM_RELTIME:
 		record_localtime(ctl, rec, &cur);
 		delta = record_count_delta(ctl, rec);
-
-		if (cur.tm_min  != ctl->lasttm.tm_min ||
+		if (cur.tm_min != ctl->lasttm.tm_min ||
 		    cur.tm_hour != ctl->lasttm.tm_hour ||
 		    cur.tm_yday != ctl->lasttm.tm_yday) {
-			if (ctl->color)
-				color_enable(DMESG_COLOR_RELTIME);
 			printf("[%s] ", short_ctime(&cur, buf, sizeof(buf)));
 		} else {
-			if (ctl->color)
-				color_enable(DMESG_COLOR_TIME);
 			if (delta < 10)
 				printf("[  %+8.06f] ", delta);
 			else
 				printf("[ %+9.06f] ", delta);
 		}
-		if (ctl->color)
-			color_disable();
 		ctl->lasttm = cur;
+		break;
+	case DMESG_TIMEFTM_TIME:
+		printf("[%5d.%06d] ", (int)rec->tv.tv_sec, (int)rec->tv.tv_usec);
+		break;
+	case DMESG_TIMEFTM_TIME_DELTA:
+		printf("[%5d.%06d <%12.06f>] ", (int)rec->tv.tv_sec,
+		       (int)rec->tv.tv_usec, record_count_delta(ctl, rec));
+		break;
+	case DMESG_TIMEFTM_ISO8601:
+		printf("%s ", iso_8601_time(ctl, rec, buf, sizeof(buf)));
+		break;
+	default:
+		abort();
 	}
 
-	/*
-	 * In syslog output the timestamp is part of the message and we don't
-	 * parse the timestamp by default. We parse the timestamp only if
-	 * --show-delta or --ctime is specified.
-	 *
-	 * In kmsg output we always parse the timesptamp, so we have to compose
-	 * the [sec.usec] string.
-	 */
-	if (ctl->method == DMESG_METHOD_KMSG &&
-	    !ctl->notime && !ctl->delta && !ctl->ctime && !ctl->reltime) {
-		if (ctl->color)
-			color_enable(DMESG_COLOR_TIME);
-		printf("[%5d.%06d] ", (int) rec->tv.tv_sec, (int) rec->tv.tv_usec);
-		if (ctl->color)
-			color_disable();
-	}
+	if (ctl->color)
+		color_disable();
 
 mesg:
 	mesg = rec->mesg;
@@ -991,6 +1013,8 @@ static int init_kmsg(struct dmesg_control *ctl)
 
 	if (!ctl->follow)
 		mode |= O_NONBLOCK;
+	else
+		setlinebuf(stdout);
 
 	ctl->kmsg = open("/dev/kmsg", mode);
 	if (ctl->kmsg < 0)
@@ -1065,7 +1089,7 @@ static int parse_kmsg_record(struct dmesg_control *ctl,
 		goto mesg;
 
 	/* C) timestamp */
-	if (ctl->notime)
+	if (is_timefmt(ctl, NONE))
 		p = skip_item(p, end, ",;");
 	else
 		p = parse_kmsg_timestamp(p, &rec->tv);
@@ -1136,24 +1160,45 @@ static int read_kmsg(struct dmesg_control *ctl)
 	return 0;
 }
 
+static int which_time_format(const char *optarg)
+{
+	if (!strcmp(optarg, "notime"))
+		return DMESG_TIMEFTM_NONE;
+	if (!strcmp(optarg, "ctime"))
+		return DMESG_TIMEFTM_CTIME;
+	if (!strcmp(optarg, "delta"))
+		return DMESG_TIMEFTM_DELTA;
+	if (!strcmp(optarg, "reltime"))
+		return DMESG_TIMEFTM_RELTIME;
+	if (!strcmp(optarg, "iso"))
+		return DMESG_TIMEFTM_ISO8601;
+	errx(EXIT_FAILURE, _("unknown time format: %s"), optarg);
+}
+
 int main(int argc, char *argv[])
 {
 	char *buf = NULL;
 	int  c, nopager = 0;
 	int  console_level = 0;
 	int  klog_rc = 0;
+	int  delta = 0;
 	ssize_t n;
 	static struct dmesg_control ctl = {
 		.filename = NULL,
 		.action = SYSLOG_ACTION_READ_ALL,
 		.method = DMESG_METHOD_KMSG,
 		.kmsg = -1,
+		.time_fmt = DMESG_TIMEFTM_TIME,
+	};
+	int colormode = UL_COLORMODE_NEVER;
+	enum {
+		OPT_TIME_FORMAT = CHAR_MAX + 1,
 	};
 
 	static const struct option longopts[] = {
 		{ "buffer-size",   required_argument, NULL, 's' },
 		{ "clear",         no_argument,	      NULL, 'C' },
-		{ "color",         no_argument,	      NULL, 'L' },
+		{ "color",         optional_argument, NULL, 'L' },
 		{ "console-level", required_argument, NULL, 'n' },
 		{ "console-off",   no_argument,       NULL, 'D' },
 		{ "console-on",    no_argument,       NULL, 'E' },
@@ -1175,14 +1220,20 @@ int main(int argc, char *argv[])
 		{ "nopager",       no_argument,       NULL, 'P' },
 		{ "userspace",     no_argument,       NULL, 'u' },
 		{ "version",       no_argument,	      NULL, 'V' },
+		{ "time-format",   required_argument, NULL, OPT_TIME_FORMAT },
 		{ NULL,	           0, NULL, 0 }
 	};
 
 	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
-		{ 'C','D','E','c','n' },	/* clear,off,on,read-clear,level*/
+		{ 'C','D','E','c','n','r' },	/* clear,off,on,read-clear,level,raw*/
 		{ 'H','r' },			/* human, raw */
 		{ 'L','r' },			/* color, raw */
 		{ 'S','w' },			/* syslog,follow */
+		{ 'T','r' },			/* ctime, raw */
+		{ 'd','r' },			/* delta, raw */
+		{ 'e','r' },			/* reltime, raw */
+		{ 'r','x' },			/* raw, decode */
+		{ 'r','t' },			/* notime, raw */
 		{ 0 }
 	};
 	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
@@ -1192,7 +1243,7 @@ int main(int argc, char *argv[])
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
-	while ((c = getopt_long(argc, argv, "CcDdEeF:f:HhkLl:n:iPrSs:TtuVwx",
+	while ((c = getopt_long(argc, argv, "CcDdEeF:f:HhkL::l:n:iPrSs:TtuVwx",
 				longopts, NULL)) != -1) {
 
 		err_exclusive_options(c, longopts, excl, excl_st);
@@ -1208,13 +1259,13 @@ int main(int argc, char *argv[])
 			ctl.action = SYSLOG_ACTION_CONSOLE_OFF;
 			break;
 		case 'd':
-			ctl.delta = 1;
+			delta = 1;
 			break;
 		case 'E':
 			ctl.action = SYSLOG_ACTION_CONSOLE_ON;
 			break;
 		case 'e':
-			ctl.reltime = 1;
+			ctl.time_fmt = DMESG_TIMEFTM_RELTIME;
 			break;
 		case 'F':
 			ctl.filename = optarg;
@@ -1227,8 +1278,8 @@ int main(int argc, char *argv[])
 				return EXIT_FAILURE;
 			break;
 		case 'H':
-			ctl.reltime = 1;
-			ctl.color = 1;
+			ctl.time_fmt = DMESG_TIMEFTM_RELTIME;
+			colormode = UL_COLORMODE_AUTO;
 			ctl.pager = 1;
 			break;
 		case 'h':
@@ -1239,7 +1290,10 @@ int main(int argc, char *argv[])
 			setbit(ctl.facilities, FAC_BASE(LOG_KERN));
 			break;
 		case 'L':
-			ctl.color = 1;
+			colormode = UL_COLORMODE_AUTO;
+			if (optarg)
+				colormode = colormode_or_err(optarg,
+						_("unsupported color mode"));
 			break;
 		case 'l':
 			ctl.fltr_lev= 1;
@@ -1267,12 +1321,11 @@ int main(int argc, char *argv[])
 				ctl.bufsize = 4096;
 			break;
 		case 'T':
-			ctl.boot_time = get_boot_time();
-			if (ctl.boot_time)
-				ctl.ctime = 1;
+			ctl.time_fmt = DMESG_TIMEFTM_CTIME;
 			break;
 		case 't':
-			ctl.notime = 1;
+			ctl.time_fmt = DMESG_TIMEFTM_NONE;
+			delta = 0;
 			break;
 		case 'u':
 			ctl.fltr_fac = 1;
@@ -1289,6 +1342,9 @@ int main(int argc, char *argv[])
 		case 'x':
 			ctl.decode = 1;
 			break;
+		case OPT_TIME_FORMAT:
+			ctl.time_fmt = which_time_format(optarg);
+			break;
 		case '?':
 		default:
 			usage(stderr);
@@ -1300,24 +1356,33 @@ int main(int argc, char *argv[])
 	if (argc > 1)
 		usage(stderr);
 
-	if (ctl.raw && (ctl.fltr_lev || ctl.fltr_fac || ctl.delta ||
-			ctl.notime || ctl.ctime || ctl.decode))
-		errx(EXIT_FAILURE, _("--raw can't be used together with level, "
-		     "facility, decode, delta, ctime or notime options"));
+	if (is_timefmt(&ctl, RELTIME) ||
+	    is_timefmt(&ctl, CTIME) ||
+	    is_timefmt(&ctl, ISO8601)) {
 
-	if (ctl.notime && (ctl.ctime || ctl.reltime))
-		errx(EXIT_FAILURE, _("--notime can't be used together with --ctime or --reltime"));
-	if (ctl.reltime && ctl.ctime)
-		errx(EXIT_FAILURE, _("--reltime can't be used together with --ctime "));
-
-	if (ctl.reltime) {
-		ctl.boot_time = get_boot_time();
-		if (!ctl.boot_time)
-			ctl.reltime = 0;
+		if (get_boot_time(&ctl.boot_time) != 0)
+			ctl.time_fmt = DMESG_TIMEFTM_NONE;
 	}
-	if (ctl.color)
-		ctl.color = colors_init() ? 1 : 0;
 
+	if (delta)
+		switch (ctl.time_fmt) {
+		case DMESG_TIMEFTM_CTIME:
+			ctl.time_fmt = DMESG_TIMEFTM_CTIME_DELTA;
+			break;
+		case DMESG_TIMEFTM_TIME:
+			ctl.time_fmt = DMESG_TIMEFTM_TIME_DELTA;
+			break;
+		case DMESG_TIMEFTM_ISO8601:
+			warnx(_("--show-delta is ignored when used together with iso8601 time format"));
+			break;
+		default:
+			ctl.time_fmt = DMESG_TIMEFTM_DELTA;
+		}
+
+
+	ctl.color = colors_init(colormode) ? 1 : 0;
+	if (ctl.follow)
+		nopager = 1;
 	ctl.pager = nopager ? 0 : ctl.pager;
 	if (ctl.pager)
 		setup_pager();
@@ -1327,6 +1392,12 @@ int main(int argc, char *argv[])
 	case SYSLOG_ACTION_READ_CLEAR:
 		if (ctl.method == DMESG_METHOD_KMSG && init_kmsg(&ctl) != 0)
 			ctl.method = DMESG_METHOD_SYSLOG;
+
+		if (ctl.raw
+		    && ctl.method != DMESG_METHOD_KMSG
+		    && (ctl.fltr_lev || ctl.fltr_fac))
+			    errx(EXIT_FAILURE, _("--raw could be used together with --level or "
+				 "--facility only when read messages from /dev/kmsg"));
 		if (ctl.pager)
 			setup_pager();
 		n = read_buffer(&ctl, &buf);
