@@ -32,7 +32,7 @@
  */
 
 /*
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+ * 1999-02-22 Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  * 2000-07-30 Per Andreas Buer <per@linpro.no> - added "q"-option
@@ -58,6 +58,7 @@
 #include <limits.h>
 #include <locale.h>
 #include <stddef.h>
+#include <poll.h>
 
 #include "closestream.h"
 #include "nls.h"
@@ -159,7 +160,7 @@ main(int argc, char **argv) {
 	sigset_t block_mask, unblock_mask;
 	struct sigaction sa;
 	int ch;
-	FILE *timingfd = stderr;
+	FILE *timingfd = NULL;
 
 	enum { FORCE_OPTION = CHAR_MAX + 1 };
 
@@ -274,9 +275,11 @@ main(int argc, char **argv) {
 			warn(_("fork failed"));
 			fail();
 		}
-		if (child)
+		if (child) {
+			if (!timingfd)
+				timingfd = fdopen(STDERR_FILENO, "w");
 			dooutput(timingfd);
-		else
+		} else
 			doshell();
 	} else {
 		sa.sa_handler = resize;
@@ -289,7 +292,7 @@ main(int argc, char **argv) {
 	return EXIT_SUCCESS;
 }
 
-void __attribute__((__noreturn__))
+void
 doinput(void) {
 	ssize_t cc;
 	char ibuf[BUFSIZ];
@@ -306,7 +309,12 @@ doinput(void) {
 			}
 		}
 		else if (cc < 0 && errno == EINTR && resized)
+		{
+			/* transmit window change information to the child */
+			ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&win);
+			ioctl(slave, TIOCSWINSZ, (char *)&win);
 			resized = 0;
+		}
 		else
 			break;
 	}
@@ -320,20 +328,20 @@ void
 finish(int dummy __attribute__ ((__unused__))) {
 	int status;
 	pid_t pid;
+	int errsv = errno;
 
 	while ((pid = wait3(&status, WNOHANG, 0)) > 0)
 		if (pid == child) {
 			childstatus = status;
 			die = 1;
 		}
+
+	errno = errsv;
 }
 
 void
 resize(int dummy __attribute__ ((__unused__))) {
 	resized = 1;
-	/* transmit window change information to the child */
-	ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&win);
-	ioctl(slave, TIOCSWINSZ, (char *)&win);
 }
 
 /*
@@ -345,16 +353,16 @@ my_strftime(char *buf, size_t len, const char *fmt, const struct tm *tm) {
 	strftime(buf, len, fmt, tm);
 }
 
-void __attribute__((__noreturn__))
+void
 dooutput(FILE *timingfd) {
 	ssize_t cc;
 	time_t tvec;
 	char obuf[BUFSIZ];
 	struct timeval tv;
 	double oldtime=time(NULL), newtime;
-	int flgs = 0;
 	ssize_t wrt;
 	ssize_t fwrt;
+	int errsv = 0;
 
 	close(STDIN_FILENO);
 #ifdef HAVE_LIBUTIL
@@ -365,25 +373,21 @@ dooutput(FILE *timingfd) {
 	fprintf(fscript, _("Script started on %s"), obuf);
 
 	do {
-		if (die && flgs == 0) {
-			/* ..child is dead, but it doesn't mean that there is
-			 * nothing in buffers.
-			 */
-			flgs = fcntl(master, F_GETFL, 0);
-			if (fcntl(master, F_SETFL, (flgs | O_NONBLOCK)) == -1)
+		if (die || errsv == EINTR) {
+			struct pollfd fds[] = {{ .fd = master, .events = POLLIN }};
+			if (poll(fds, 1, 50) <= 0)
 				break;
 		}
-		if (tflg)
-			gettimeofday(&tv, NULL);
 
 		errno = 0;
 		cc = read(master, obuf, sizeof (obuf));
+		errsv = errno;
 
-		if (die && errno == EINTR && cc <= 0)
-			/* read() has been interrupted by SIGCHLD, try it again
-			 * with O_NONBLOCK
-			 */
-			continue;
+		if (tflg)
+			gettimeofday(&tv, NULL);
+
+		if (errsv == EINTR && cc <= 0)
+			continue;	/* try it again */
 		if (cc <= 0)
 			break;
 		if (tflg) {
@@ -391,28 +395,29 @@ dooutput(FILE *timingfd) {
 			fprintf(timingfd, "%f %zd\n", newtime - oldtime, cc);
 			oldtime = newtime;
 		}
-		wrt = write(STDOUT_FILENO, obuf, cc);
-		if (wrt < 0) {
-			warn (_("write failed"));
-			fail();
-		}
 		fwrt = fwrite(obuf, 1, cc, fscript);
 		if (fwrt < cc) {
 			warn (_("cannot write script file"));
 			fail();
 		}
-		if (fflg)
+		if (fflg) {
 			fflush(fscript);
+			if (tflg)
+				fflush(timingfd);
+		}
+		wrt = write(STDOUT_FILENO, obuf, cc);
+		if (wrt < 0) {
+			warn (_("write failed"));
+			fail();
+		}
 	} while(1);
 
-	if (flgs)
-		fcntl(master, F_SETFL, flgs);
 	if (close_stream(timingfd) != 0)
 		errx(EXIT_FAILURE, _("write error"));
 	done();
 }
 
-void __attribute__((__noreturn__))
+void
 doshell(void) {
 	char *shname;
 
@@ -443,11 +448,17 @@ doshell(void) {
 	 */
 	signal(SIGTERM, SIG_DFL);
 
-	if (cflg)
-		execl(shell, shname, "-c", cflg, NULL);
-	else
-		execl(shell, shname, "-i", NULL);
-
+	if (access(shell, X_OK) == 0) {
+		if (cflg)
+			execl(shell, shname, "-c", cflg, NULL);
+		else
+			execl(shell, shname, "-i", NULL);
+	} else {
+		if (cflg)
+			execlp(shname, "-c", cflg, NULL);
+		else
+			execlp(shname, "-i", NULL);
+	}
 	warn(_("failed to execute %s"), shell);
 	fail();
 }
@@ -462,7 +473,7 @@ fixtty(void) {
 	tcsetattr(STDIN_FILENO, TCSANOW, &rtt);
 }
 
-void __attribute__((__noreturn__))
+void
 fail(void) {
 
 	kill(0, SIGTERM);
